// calc2.cc
// demonstration of external constants, variables, pointers,
// functions using


#include <iostream>
#include <vector>
#include <signal.h>

#include <forint/forint.hh>

void err_quit(void) { cerr << "\n-----ERROR QUIT-----" << endl; exit(1); }

// extern functions
double f0(void)   { return M_PI; }

double f1(double a)
{
  if (a < 0) return a;
  if (a == 0) return 1.;
  return (long long)a * f1((long long)(a-1));
}

double f2(double a, double b)
{ return (a) ? a * b : b; }

double f3(void* a, void* b)
{ return ((double*)a)[0] * ((double*)b)[1]; }

double f4(double i, void* a)
{ return (i > 7 && i < 0) ? 0. : ((double*)a) [long(i)]; }

class classA {
public:
  double d;
classA (void) { d = M_PI; }
};

class classB {
public:
  int i;
classB (void) { i = 100; }
};

class classC {
public:
  int i;
classC (void) { i = 100; }
};

double f5(void* v1, void* v2)
{
  classA* a;
  classB* b;
  if ((a=static_cast<classA*>(v1)))
    cout << "  a: " << a->d << endl;
  else {
    cerr << "\nError in dynamic_cast<classA*>" << endl;
    err_quit();
  }
  if ((b=static_cast<classB*>(v2)))
    cout << "  b: " << b->i << endl;
  else {
    cerr << "\nError in dynamic_cast<classB*>" << endl;
    err_quit();
  }
  return 0;
}

bool* user_break = 0;

void
handler(int sig)
{
  switch (sig) {
  case SIGINT:
    cerr << "\n*** Use 'q' to quit ***" << endl;
    if (user_break)
      *user_break = true;
    if (signal(SIGINT,handler) == SIG_ERR)
      cerr << "\n*** calc3: signal(SIGINT) error ***\n" << endl;
    break;
  case SIGFPE:
    cerr << "\n*** Floating point exception ***" << endl;
    if (signal(SIGFPE,handler) == SIG_ERR)
      cerr << "\n*** calc3: signal(SIGFPE) error ***\n" << endl;
    break;
  case SIGSEGV:
    cerr << "\n*** Invalid memory reference ***" << endl;
    if (signal(SIGSEGV,handler) == SIG_ERR)
      cerr << "\n*** calc3: signal(SIGSEGV) error ***\n" << endl;
    break;
  default:;
  }
}

main()
{
  using namespace FInt;
  cout << "this programm demonstrates accessability of external:\n\n"
       << "  constants:  double c1, c2;\n"
       << "  variables:  double v1, v2;\n"
       << "  pointers :  double A[2][3], **C;\n"
       << "              classA* a;\n"
       << "              classB* b;\n"
       << "              classC* c;\n"
       << "  functions:  double F0(void);\n"
       << "              double F1(double);\n"
       << "              double F2(double, double);\n"
       << "              double F3(double*, double*);\n"
       << "              double F4(double, double*);\n"
       << "              double F5(classA*, classB*);\n\n"
       << "To quit enter 'q'\n";

  string expr;

// CLI object
  mssg cli;

// forint object
  forint calc(cli);

  fi_version_t version;
  calc.getVersion(version);
  cout << "library name:  " << version.name << endl
       << "major version: " << version.major_version << endl
       << "minor version: " << version.minor_version << endl
       << "release:       " << version.release << endl;

// get pointer to break flag (to exit from an endless loop)
  user_break = calc.getPBrk();

  if (signal(SIGINT,SIG_IGN) != SIG_IGN)
    if (signal(SIGINT,handler) == SIG_ERR)
      cerr << "calc3: signal(SIGINT) error.\n" << endl;
  if (signal(SIGFPE,handler) == SIG_ERR)
    cerr << "calc3: signal(SIGFPE) error.\n" << endl;
  if (signal(SIGSEGV,handler) == SIG_ERR)
    cerr << "calc3: signal(SIGSEGV) error.\n" << endl;

// put external constants
  const double const1=100;
// pseudo-constant, programmer can change it, user can't
  double const2 = 200;
  if (calc.putConst(const1, "c1"))
    err_quit();
  if (calc.putConst(const2, "c2"))
    err_quit();

// put external variables
  double var1=1, var2=2;
  if (calc.putVar(var1,"v1"))
    err_quit();
  if (calc.putVar(var2,"v2"))
    err_quit();

// put a static array
  double a [2][3];
// a[i][j] - array's element 
  for (int i=0; i<2; i++)
    for (int j=0; j<3; j++)
      a[i][j] = i*3+j;
  vector<int> dims_a(2); 
  dims_a[0] = 2; 
  dims_a[1] = 3; 
  if (calc.putPoin(a, "A", "double", &dims_a))
    err_quit();

// put a dynamic array
  double** pc  = new double* [2];
  double*  c = new double [2 * 3]; 
  pc[0] = c; 
  pc[1] = c + 3;
// pc[i][j] - array's element 
  for (int i=0; i<2; i++)
    for (int j=0; j<3; j++)
      pc[i][j] = 10+i*3+j;
  vector<int> dims_c(2); 
  dims_c[0] = 2; 
  dims_c[1] = 3; 
  if (calc.putPoin(c, "C", "double", &dims_c))
    err_quit();

// put external functions
  if (calc.putFunc0 (f0, "F0"))
    err_quit();
  if (calc.putFunc1 (f1, "F1"))
    err_quit();
  if (calc.putFunc2D(f2, "F2"))
    err_quit();
  if (calc.putFunc2P(f3, "F3","double*","double*"))
    err_quit();
  if (calc.putFunc2 (f4, "F4","double*"))
    err_quit();

  classA obj_a;
  classB obj_b;
  classC obj_c;
  if (calc.putPoin(&obj_a, "a", "classA*"))
    err_quit();
  if (calc.putPoin(&obj_b, "b", "classB*"))
    err_quit();
  if (calc.putPoin(&obj_c, "c", "classC*"))
    err_quit();
  if (calc.putFunc2P (f5, "F5","classA*","classB*"))
    err_quit();

  double result;
  cout.precision(15);

  while (1) {

    cout << "calc2> " << flush;
    getline(cin,expr);

    if (expr == "q") break;

// parse
    if (!calc.parse((char*)expr.c_str(), // string to parse
		    NEW_FORM_MODE, // parsing mode
		    false,        // switch off forint built-in variables
		    false))       // switch off 'rl' and 'self'
      {
	result = calc.calculate(SVM_MODE);
// if no errors occured calculate and print the result
	if (isinf(result)<0) {
	  cerr << "\n*** Negative  infinity ***" << endl;
	  continue;
	}
	if (isinf(result)>0) {
	  cerr << "\n*** Positive  infinity ***" << endl;
	  continue;
	}
	if (isnan(result)) {
	  cerr << "\n*** Not-a-number ***" << endl;
	  continue;
	}
	cout << "     " << result;
      }
    cout << endl;
  }
}
